searchData={"items":[{"type":"module","title":"Torus","doc":"Torus is a plug-and-play Elixir library that seamlessly integrates PostgreSQL's search into Ecto, streamlining the construction of advanced search queries.","ref":"Torus.html"},{"type":"module","title":"Usage - Torus","doc":"The package can be installed by adding `torus` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:torus, \"~> 0.5\"}\n  ]\nend\n```\n\nThen, in any query, you can (for example) add a prefixed full-text search:\n\n```elixir\nimport Torus\n# ...\n\nPost\n# ... your complex query\n|> Torus.full_text([p], [p.title, p.body], \"uncove hogwar\")\n|> select([p], p.title)\n|> Repo.all()\n[\"Uncovered hogwarts\"]\n```\n\nSee [`full_text/5`](https://hexdocs.pm/torus/Torus.html#full_text/5) for more details.","ref":"Torus.html#module-usage"},{"type":"module","title":"6 types of search: - Torus","doc":"1. **Pattern matching**: Searches for a specific pattern in a string.\n\n   ```elixir\n   iex> insert_posts!([\"Wand\", \"Magic wand\", \"Owl\"])\n   ...> Post\n   ...> |> Torus.ilike([p], [p.title], \"wan%\")\n   ...> |> select([p], p.title)\n   ...> |> Repo.all()\n   [\"Wand\"]\n   ```\n\n   See [`like/5`](https://hexdocs.pm/torus/Torus.html#like/5), [`ilike/5`](https://hexdocs.pm/torus/Torus.html#ilike/5), and [`similar_to/5`](https://hexdocs.pm/torus/Torus.html#similar_to/5) for more details.\n\n1. **Similarity:** Searches for records that closely match the input text, often using trigram or Levenshtein distance. Ideal for fuzzy matching and catching typos in short text fields.\n\n   ```elixir\n   iex> insert_posts!([\"Hogwarts Secrets\", \"Quidditch Fever\", \"Hogwart’s Secret\"])\n   ...> Post\n   ...> |> Torus.similarity([p], [p.title], \"hoggwarrds\")\n   ...> |> limit(2)\n   ...> |> select([p], p.title)\n   ...> |> Repo.all()\n   [\"Hogwarts Secrets\", \"Hogwart’s Secret\"]\n   ```\n\n   See [`similarity/5`](https://hexdocs.pm/torus/Torus.html#similarity/5) for more details.\n\n1. **Text Search Vectors**: Uses term-document matrix vectors for **full-text search**, enabling efficient querying and ranking based on term frequency. - [PostgreSQL: Full Text Search](https://www.postgresql.org/docs/current/textsearch.html). Is great for large datasets to quickly return relevant results.\n\n   ```elixir\n   iex> insert_post!(title: \"Hogwarts Shocker\", body: \"A spell disrupts the Quidditch Cup.\")\n   ...> insert_post!(title: \"Diagon Bombshell\", body: \"Secrets uncovered in the heart of Hogwarts.\")\n   ...> insert_post!(title: \"Completely unrelated\", body: \"No magic here!\")\n   ...> Post\n   ...> |> Torus.full_text([p], [p.title, p.body], \"uncov hogwar\")\n   ...> |> select([p], p.title)\n   ...> |> Repo.all()\n   [\"Diagon Bombshell\"]\n   ```\n\n   See [`full_text/5`](https://hexdocs.pm/torus/Torus.html#full_text/5) for more details.\n\n1. **Semantic Search**: Understands the contextual meaning of queries to match and retrieve related content utilizing natural language processing. Read more about semantic search in [Semantic search with Torus guide](/guides/semantic_search.md).\n\n   ```elixir\n   insert_post!(title: \"Hogwarts Shocker\", body: \"A spell disrupts the Quidditch Cup.\")\n   insert_post!(title: \"Diagon Bombshell\", body: \"Secrets uncovered in the heart of Hogwarts.\")\n   insert_post!(title: \"Completely unrelated\", body: \"No magic here!\")\n\n   embedding_vector = Torus.to_vector(\"A magic school in the UK\")\n\n   Post\n   |> Torus.semantic([p], p.embedding, embedding_vector)\n   |> select([p], p.title)\n   |> Repo.all()\n   [\"Diagon Bombshell\"]\n   ```\n\n   See [`semantic/5`](https://hexdocs.pm/torus/Torus.html#semantic/5) for more details.\n\n1. **Hybrid Search**: Combines multiple search techniques (e.g., keyword and semantic) to leverage their strengths for more accurate results.\n\n   Will be added soon.\n\n1. **3rd Party Engines/Providers**: Utilizes external services or software specifically designed for optimized and scalable search capabilities, such as Elasticsearch or Algolia.","ref":"Torus.html#module-6-types-of-search"},{"type":"module","title":"Optimizations and relevance - Torus","doc":"Torus is designed to be as efficient and relevant as possible from the start. But handling large datasets and complex search queries tends to be tricky. The best way to combine these two to achieve the best result is to:\n\n1. Create a query that returns as relevant results as possible (by tweaking the options of search function). If there is any option missing - feel free to open an issue/contribute back with it, or implement it manually using fragments.\n2. Test its performance on real production data - maybe it's good enough already?\n3. If it's not:\n   - See optimization sections for your search type in `Torus` docs\n   - Inspect your query using [`Torus.QueryInspector.tap_substituted_sql/3`](https://hexdocs.pm/torus/Torus.QueryInspector.html#tap_substituted_sql/3) or [`Torus.QueryInspector.tap_explain_analyze/3`](https://hexdocs.pm/torus/Torus.QueryInspector.html#tap_explain_analyze/3)\n   - According to the above SQL - add indexes for the queried rows/vectors","ref":"Torus.html#module-optimizations-and-relevance"},{"type":"module","title":"Debugging your queries - Torus","doc":"Torus offers a few helpers to debug, explain, and analyze your queries before using them on production. See `Torus.QueryInspector` for more details.","ref":"Torus.html#module-debugging-your-queries"},{"type":"module","title":"Torus support - Torus","doc":"For now, Torus supports pattern match, similarity, full-text, and semantic search, with plans to expand support further. These docs will be updated with more examples on which search type to choose and how to make them more performant (by adding indexes or using specific functions).","ref":"Torus.html#module-torus-support"},{"type":"function","title":"Torus.embedding_model/1","doc":"Calls the specified embedding module's `embedding_model/1` function to retrieve the model name.\n\nSee [Semantic search guide](semantic_search.html) for more info.","ref":"Torus.html#embedding_model/1"},{"type":"macro","title":"Torus.full_text/5","doc":"Full text search with rank ordering. Accepts a list of columns to search in. A list of columns\ncan either be a text or `tsvector` type. If `tsvector`s are passed make sure to set\n`stored: true`.\n\nCleans the term, so it can be input directly by the user. The default preset of\nsettings is optimal for most cases.\n\nFull Text Searching (or just text search) provides the capability to identify\nnatural-language documents that satisfy a query, and optionally to sort them by\nrelevance to the query. The most common type of search is to find all documents\ncontaining given query terms and return them in order of their similarity to the query.\nNotions of query and similarity are very flexible and depend on the specific application.\nThe simplest search considers query as a set of words and similarity as the frequency of\nquery words in the document. Read more in [PostgreSQL Full Text Search docs](https://postgresql.org/docs/current/interactive/textsearch-intro.html).","ref":"Torus.html#full_text/5"},{"type":"macro","title":"Options - Torus.full_text/5","doc":"* `:prefix_search` - whether to apply prefix search.\n    - `true` (default) - the term is treated as a prefix\n    - `false` - only counts full-word matches\n  * `:stored` - whether to use stored tsvector or not.\n    - `false` (default) - columns (or expressions) passed as qualifiers are of type `text`\n    - `true` - columns (or expressions) passed as qualifiers are **tsvectors**\n  * `:language` - language used for the search. Defaults to `\"english\"`.\n  * `:term_function` - function used to convert the term to `ts_query`. Can be one of:\n    - `:websearch_to_tsquery` (default) - converts term to a tsquery, normalizing\n    words according to the specified or default configuration. Quoted word sequences\n    are converted to phrase tests. The word “or” is understood as producing an OR\n    operator, and a dash produces a NOT operator; other punctuation is ignored. This\n    approximates the behavior of some common web search tools.\n    - `:plainto_tsquery` - converts term to a tsquery, normalizing words according\n    to the specified or default configuration. Any punctuation in the string is\n    ignored (it does not determine query operators). The resulting query matches\n    documents containing all non-stopwords in the term.\n    - `:phraseto_tsquery` - converts term to a tsquery, normalizing words according\n    to the specified or default configuration. Any punctuation in the string is\n    ignored (it does not determine query operators). The resulting query matches\n    phrases containing all non-stopwords in the text.\n  * `:rank_function` - function used to rank the results.\n    - `:ts_rank_cd` (default) - computes a score showing how well the vector matches\n    the query, using a cover density algorithm. See [Ranking Search Results](https://postgresql.org/docs/current/interactive/textsearch-controls.html#TEXTSEARCH-RANKING) for more details.\n    - `:ts_rank` - computes a score showing how well the vector matches the query.\n  * `:rank_weights` - a list of weights for each column. Defaults to `[:A, :B, :C, :D]`.\n  The length of weights (if provided) should be the same as the length of the columns we search for.\n  A single weight can be either a string or an atom. Possible values are:\n    - `:A` - 1.0\n    - `:B` - 0.4\n    - `:C` - 0.2\n    - `:D` - 0.1\n  * `:rank_normalization` - a string that specifies whether and how a document's\n  length should impact its rank. The integer option controls several behaviors, so\n  it is a bit mask: you can specify one or more behaviors using `|` (for example, `2|4`).\n    - `0` (default for `ts_rank`) - ignores the document length\n    - `1`  - divides the rank by 1 + the logarithm of the document length\n    - `2`  - divides the rank by the document length\n    - `4` (default for `ts_rank_cs`)  - divides the rank by the mean harmonic\n    distance between extents (this is implemented only by `ts_rank_cd`)\n    - `8`  - divides the rank by the number of unique words in document\n    - `16` -  divides the rank by 1 + the logarithm of the number of unique words in\n    document\n    - `32` - divides the rank by itself + 1\n  * `:order` - describes the ordering of the results. Possible values are\n    - `:desc` (default) - orders the results by similarity rank in descending order.\n    - `:asc` - orders the results by similarity rank in ascending order.\n    - `:none` - doesn't apply ordering at all.\n  * `:filter_type` - filter type\n    - `:or` (default) - uses `OR` operator to combine different column matches.\n    Selecting this option means that the search term won't match across columns.\n    - `:concat` - joins the columns into a single tsvector and searches for the\n    term in the concatenated string containing all columns.\n    - `:none` - doesn't apply any filtering and returns all results.\n  * `empty_return` - whether to return all results when the search term is empty.\n    - `true` (default) - returns all results when the search term is empty.\n    - `false` - returns an empty list when the search term is empty.\n  * `:coalesce` - when joining columns via `:concat` option, adds a\n  `COALESCE` function to handle NULL values. Choose true when you can't guarantee\n  that all columns are non-null.\n    - `true` (default)- adds `COALESCE`\n    - `false` - doesn't add `COALESCE` function to the query. Choose this when you're\n    using `filter_type: :concat` and can guarantee that all columns are non-null.","ref":"Torus.html#full_text/5-options"},{"type":"macro","title":"Example usage - Torus.full_text/5","doc":"iex> insert_post!(title: \"Hogwarts Shocker\", body: \"A spell disrupts the Quidditch Cup.\")\n    ...> insert_post!(title: \"Diagon Bombshell\", body: \"Secrets uncovered in the heart of Hogwarts.\")\n    ...> insert_post!(title: \"Completely unrelated\", body: \"No magic here!\")\n    ...>  Post\n    ...> |> Torus.full_text([p], [p.title, p.body], \"uncov hogwar\")\n    ...> |> select([p], p.title)\n    ...> |> Repo.all()\n    [\"Diagon Bombshell\"]","ref":"Torus.html#full_text/5-example-usage"},{"type":"macro","title":"Optimizations - Torus.full_text/5","doc":"- Store precomputed tsvector in a separate column, add a GIN index to it, and use\n  `stored: true`.\n\n  - Add a GIN tsvector index on the column(s) you search in.\n  Use `Torus.QueryInspector.tap_sql/2` on your query (with all the options passed) to see the exact search string and add an index to it. For example for nullable title, the GIN index could look like:\n\n    ```sql\n    CREATE INDEX index_gin_posts_title\n    ON posts USING GIN (to_tsvector('english', COALESCE(title, '')));\n    ```","ref":"Torus.html#full_text/5-optimizations"},{"type":"macro","title":"Torus.ilike/5","doc":"Case-insensitive pattern matching search using\n[PostgreSQL `ILIKE`](https://www.postgresql.org/docs/current/functions-matching.html#FUNCTIONS-LIKE) operator.\n\n> #### Warning {: .neutral}\n>\n> Doesn't clean the term, so it needs to be sanitized before being passed in. See\n[LIKE-injections](https://githubengineering.com/like-injection/).","ref":"Torus.html#ilike/5"},{"type":"macro","title":"Examples - Torus.ilike/5","doc":"iex> insert_posts!(titles: [\"Wand\", \"Magic wand\", \"Owl\"])\n    ...> Post\n    ...> |> Torus.ilike([p], [p.title], \"wan%\")\n    ...> |> select([p], p.title)\n    ...> |> Repo.all()\n    [\"Wand\"]\n\n    iex> insert_posts!([%{title: \"hogwarts\", body: nil}, %{title: nil, body: \"HOGWARTS\"}])\n    ...> Post\n    ...> |> Torus.ilike([p], [p.title, p.body], \"%OGWART%\")\n    ...> |> select([p], %{title: p.title, body: p.body})\n    ...> |> order_by(:id)\n    ...> |> Repo.all()\n    [%{title: \"hogwarts\", body: nil}, %{title: nil, body: \"HOGWARTS\"}]\n\n    iex> insert_post!(title: \"MaGiC\")\n    ...> Post\n    ...> |> Torus.ilike([p], p.title, \"magi%\")\n    ...> |> select([p], p.title)\n    ...> |> Repo.all()\n    [\"MaGiC\"]","ref":"Torus.html#ilike/5-examples"},{"type":"macro","title":"Optimizations - Torus.ilike/5","doc":"See `like/5` optimization section for more details.","ref":"Torus.html#ilike/5-optimizations"},{"type":"macro","title":"Torus.like/5","doc":"Case-sensitive pattern matching search using [PostgreSQL `LIKE`](https://www.postgresql.org/docs/current/functions-matching.html#FUNCTIONS-LIKE) operator.\n\n> #### Warning {: .neutral}\n>\n> Doesn't clean the term, so it needs to be sanitized before being passed in. See\n[LIKE-injections](https://githubengineering.com/like-injection/).","ref":"Torus.html#like/5"},{"type":"macro","title":"Examples - Torus.like/5","doc":"iex> insert_posts!([%{title: \"hogwarts\", body: nil}, %{title: nil, body: \"HOGWARTS\"}])\n    ...> Post\n    ...> |> Torus.like([p], [p.title, p.body], \"%OGWART%\")\n    ...> |> select([p], p.body)\n    ...> |> Repo.all()\n    [\"HOGWARTS\"]","ref":"Torus.html#like/5-examples"},{"type":"macro","title":"Optimizations - Torus.like/5","doc":"- `like/5` is case-sensitive, so it can take advantage of B-tree indexes when there\nis no wildcard (%) at the beginning of the search term, prefer it over `ilike/5` if\npossible.\n\n  Adding a B-tree index:\n\n    ```sql\n    CREATE INDEX index_posts_on_title ON posts (title);\n    ```\n\n- Use `GIN` or `GiST` Index with `pg_trgm` extension for LIKE and ILIKE.\n\n  When searching for substrings (%word%), B-tree indexes won't help. Instead,\n  use trigram indexing (`pg_trgm` extension):\n\n    ```sql\n    CREATE EXTENSION IF NOT EXISTS pg_trgm;\n    CREATE INDEX posts_title_trgm_idx ON posts USING GIN (title gin_trgm_ops);\n    ```\n\n- If using prefix search, convert data to lowercase and use B-tree index for\n  case-insensitive search:\n\n    ```sql\n    ALTER TABLE posts ADD COLUMN title_lower TEXT GENERATED ALWAYS AS (LOWER(title)) STORED;\n    CREATE INDEX index_posts_on_title ON posts (title_lower);\n    ```\n\n    ```elixir\n    Torus.like([p], [p.title_lower], \"hogwarts%\")\n    ```\n\n- Use full-text search for large text fields, see `full_text/5` for more\ndetails.","ref":"Torus.html#like/5-optimizations"},{"type":"function","title":"Torus.sanitize/1","doc":"Removes all like/ilike special characters from the term, so it can be used in further pattern-match searches.","ref":"Torus.html#sanitize/1"},{"type":"macro","title":"Torus.semantic/5","doc":"Semantic search using pgvector extension to compare vectors. See [Semantic search guide](semantic_search.html) for more info.","ref":"Torus.html#semantic/5"},{"type":"macro","title":"Options - Torus.semantic/5","doc":"* `:distance` - a way to calculate the distance between the vectors. Can be one of:\n    - `:l2_distance` (default) - L2 distance\n    - `:max_inner_product` - negative inner product\n    - `:cosine_distance` - cosine distance\n    - `:l1_distance` - L1 distance\n    - `:hamming_distance` - Hamming distance\n    - `:jaccard_distance` - Jaccard distance\n  * `:order` - describes the ordering of the results. Possible values are\n    - `:asc` (default) - orders the results by distance in ascending order. 0 distance means that the vectors are the same meaning the the terms are equal. The closer the vectors - more aligned are the terms.\n    - `:desc` - orders the results by distance in descending order.\n    - `:none` - doesn't apply ordering at all.\n  * `:pre_filter` - a positive float that is passed directly to the query to pre-filter the results.\n    - `:none` (default) - no pre-filtering is done.\n    - `float` - pre-filters the results before applying the order. The results with vectors distance below the pre-filter value are returned.","ref":"Torus.html#semantic/5-options"},{"type":"macro","title":"Examples - Torus.semantic/5","doc":"def search(term) do\n      search_vector = Torus.to_vector(term)\n\n      Post\n      |> Torus.semantic([p], p.embedding, search_vector)\n      |> Repo.all()\n    end","ref":"Torus.html#semantic/5-examples"},{"type":"macro","title":"Optimizations - Torus.semantic/5","doc":"- Use `pre_filter` to pre-filter the results before applying the order. This would significantly reduce the number of rows to order.\n- Index embeddings column:\n\n    - HNSW (Hierarchical Navigable Small World) - High-accuracy Approximate Nearest\n    Neighbor\n      ```sql\n      CREATE INDEX ON embeddings USING hnsw (embedding vector_l2_ops) WITH (m = 16, ef_construction = 200);\n      ```\n\n    - IVFFlat Index (Approximate Nearest Neighbor) with different similarity\n    functions. Prior to index creation, it's recommended to have some real data in place, so the quality of clusters is better.\n      - Cosine Similarity\n        ```sql\n        CREATE INDEX ON embeddings USING ivfflat (embedding vector_cosine_ops);\n        ```\n      - L2 Distance\n        ```sql\n        CREATE INDEX ON embeddings USING ivfflat (embedding vector_l2_ops);\n        ```\n      - Inner Product\n        ```sql\n        CREATE INDEX ON embeddings USING ivfflat (embedding vector_ip_ops);\n        ```","ref":"Torus.html#semantic/5-optimizations"},{"type":"macro","title":"Torus.similar_to/5","doc":"Similar to `like/5`, except that it interprets the pattern using the SQL standard's\ndefinition of a regular expression. SQL regular expressions are a curious cross between\nLIKE notation and common (POSIX) regular expression notation. See\n[PostgreSQL `SIMILAR TO`](https://postgresql.org/docs/current/interactive/functions-matching.html?fts_query=ilike#FUNCTIONS-SIMILARTO-REGEXP)","ref":"Torus.html#similar_to/5"},{"type":"macro","title":"Examples - Torus.similar_to/5","doc":"iex> insert_post!(body: \"abc\")\n    ...> Post\n    ...> |> Torus.similar_to([p], [p.title, p.body], \"%(b|d)%\")\n    ...> |> select([p], p.body)\n    ...> |> Repo.all()\n    [\"abc\"]","ref":"Torus.html#similar_to/5-examples"},{"type":"macro","title":"Optimizations - Torus.similar_to/5","doc":"- If regex is needed, use POSIX regex with `~` or `~*` operators since they _may_\nleverage GIN or GiST indexes in some cases. These operators will be introduced later on.\n- Use `ilike/5` or `like/5` when possible, `similar_to/5` almost always does full table scans\n- Filter and limit the result set as much as possible before calling `similar_to/5`","ref":"Torus.html#similar_to/5-optimizations"},{"type":"macro","title":"Torus.similarity/5","doc":"Searches for records that closely match the input text using trigram distance. Ideal for fuzzy matching and catching typos in short text fields.\n\nImplemented using case-insensitive similarity search using [PostgreSQL similarity functions](https://postgresql.org/docs/current/interactive/pgtrgm.html#PGTRGM-FUNCS-OPS).\n\n> #### Warning {: .neutral}\n>\n> You need to have `pg_trgm` extension installed.\n> ```elixir\n> defmodule YourApp.Repo.Migrations.CreatePgTrgmExtension do\n>   use Ecto.Migration\n>\n>   def change do\n>     execute \"CREATE EXTENSION IF NOT EXISTS pg_trgm\", \"DROP EXTENSION IF EXISTS pg_trgm\"\n>   end\n> end\n> ```","ref":"Torus.html#similarity/5"},{"type":"macro","title":"Options - Torus.similarity/5","doc":"* `:type` - similarity type. Possible options are:\n    - `:word_similarity` (default) - uses `pg_trgm` `word_similarity` function. Use\n    it if you're dealing with sentences and you don't want the length of the\n    strings to affect the search result.\n    - `:strict_word_similarity` - uses `strict_word_similarity` function.\n    Prioritizes full matches, forces extent boundaries to match word boundaries.\n    Since we don't have cross-word trigrams, this function actually returns\n    greatest similarity between first string and any continuous extent of words of\n    the second string.\n    - `:similarity` - uses `similarity` function. Compares the whole set of trigrams\n    instead of an ordered subset. Use this when you search for the exact phrase or a string,\n    not a word/phrase in a sentence/longer text.\n  * `:order` - describes the ordering of the results. Possible values are\n    - `:desc` (default) - orders the results by similarity rank in descending order.\n    - `:asc` - orders the results by similarity rank in ascending order.\n    - `:none` - doesn't apply ordering and returns\n  * `:pre_filter` - whether or not to pre-filter the results:\n    - `false` (default) - omits pre-filtering and returns all results.\n    - `true` -  before applying the order, pre filters (using boolean\n  operators which potentially use GIN indexes) the result set. The results above\n  `pg_trgm.{type}_threshold` are returned. It is advised to set the corresponding value to `0.3` so that more relevant results are returned.\n  For example, for `word_similarity`, we'd run `SET pg_trgm.word_similarity_threshold = 0.3;`.","ref":"Torus.html#similarity/5-options"},{"type":"macro","title":"Examples - Torus.similarity/5","doc":"iex> insert_post!(title: \"Hogwarts Shocker\", body: \"A spell disrupts the Quidditch Cup.\")\n    ...> insert_post!(title: \"Diagon Bombshell\", body: \"Secrets uncovered in the heart of Hogwarts.\")\n    ...> insert_post!(title: \"Completely unrelated\", body: \"No magic here!\")\n    ...>  Post\n    ...> |> Torus.similarity([p], [p.title, p.body], \"Diagon Bombshell\")\n    ...> |> limit(1)\n    ...> |> select([p], p.title)\n    ...> |> Repo.all()\n    [\"Diagon Bombshell\"]\n\n    iex> insert_posts!([\"Wand\", \"Owl\", \"What an amazing cloak\"])\n    ...> Post\n    ...> |> Torus.similarity([p], [p.title], \"owls\", pre_filter: true)\n    ...> |> select([p], p.title)\n    ...> |> Repo.all()\n    [\"Owl\"]","ref":"Torus.html#similarity/5-examples"},{"type":"macro","title":"Optimizations - Torus.similarity/5","doc":"- Use `pre_filter: true` to pre-filter the results before applying the order.\nThis would significantly reduce the number of rows to order. The pre-filtering\nphase uses different (boolean) similarity operators which more actively leverage\nGIN indexes.\n- Use `order: :none` argument if you don't care about the order of the results.\nThe query will return all results that are above the similarity threshold, which\nyou can set globally via `SET pg_trgm.{type}_threshold = 0.3;`, replacing type\nwith your `type` option (e.g. `word_similarity_threshold`).\n- When `order: :desc` (default) and the limit is not set, the query will do a full\ntable scan, so it's recommended to manually limit the results (by applying `where`\nor `limit` clauses to filter the rows as much as possible).","ref":"Torus.html#similarity/5-optimizations"},{"type":"macro","title":"Adding an index - Torus.similarity/5","doc":"```sql\n-- If you haven't created it yet\nCREATE EXTENSION IF NOT EXISTS pg_trgm;\n\nCREATE INDEX index_posts_on_title ON posts USING GIN (title gin_trgm_ops);\n```","ref":"Torus.html#similarity/5-adding-an-index"},{"type":"macro","title":"Torus.substring/3","doc":"The substring function with three parameters provides extraction of a substring\nthat matches an SQL regular expression pattern. The function can be written\naccording to standard SQL syntax:\n\n```sql\nsubstring('foobar' similar '%#\"o_b#\"%' escape '#')   oob\nsubstring('foobar' similar '#\"o_b#\"%' escape '#')    NULL\n```","ref":"Torus.html#substring/3"},{"type":"macro","title":"Examples - Torus.substring/3","doc":"```elixir\ninsert_post!(title: \"Hello123World\")\nPost |> select([p], substring(p.title, \"[0-9]+\", \"#\")) |> Repo.all()\n[\"123\"]\n```","ref":"Torus.html#substring/3-examples"},{"type":"function","title":"Torus.to_vector/2","doc":"Same as `to_vectors/2`, but returns the first vector from the list.","ref":"Torus.html#to_vector/2"},{"type":"function","title":"Torus.to_vectors/2","doc":"Takes a list of terms (binaries) and embedding module's specific options and passes them to `embedding_module` `generate/2` function.\n\nConfigure `embedding_module` either in `config.exs`:\n\n      config :torus, :embedding_module, Torus.Embeddings.HuggingFace\n\nor pass `embedding_module` as an option to `to_vectors/2` function. Options always\nhave greater priority than the config.\n\nSee [Semantic search guide](semantic_search.html) for more info.","ref":"Torus.html#to_vectors/2"},{"type":"module","title":"Torus.QueryInspector","doc":"Helpers to debug your SQL queries. You can use this module both while creating the\nqueries and directly in your running production shell once deployed, so that the\nexplain analyze returns more accurate results.","ref":"Torus.QueryInspector.html"},{"type":"function","title":"Torus.QueryInspector.explain_analyze/3","doc":"Returns explain analyze on the query.\n\n**Runs the query!**","ref":"Torus.QueryInspector.html#explain_analyze/3"},{"type":"function","title":"Torus.QueryInspector.substituted_sql/3","doc":"Substitutes the params into the SQL.","ref":"Torus.QueryInspector.html#substituted_sql/3"},{"type":"function","title":"Examples - Torus.QueryInspector.substituted_sql/3","doc":"iex> Post |> where(id: 1) |> Torus.QueryInspector.substituted_sql(Torus.Test.Repo)\n    \"SELECT p0.\"id\", p0.\"title\", p0.\"body\", p0.\"author_id\" FROM \"posts\" AS p0 WHERE (p0.\"id\" = 1)\"","ref":"Torus.QueryInspector.html#substituted_sql/3-examples"},{"type":"function","title":"Torus.QueryInspector.tap_explain_analyze/3","doc":"Runs explain analyze on the query and prints it to the console. Returns the query.\n\n**Runs the query!**","ref":"Torus.QueryInspector.html#tap_explain_analyze/3"},{"type":"function","title":"Torus.QueryInspector.tap_sql/3","doc":"Converts the query to SQL and prints it to the console. Returns the query.","ref":"Torus.QueryInspector.html#tap_sql/3"},{"type":"function","title":"Torus.QueryInspector.tap_substituted_sql/3","doc":"Substitutes the parameters in the query and prints the SQL to the console. Returns the query.\nThe SQL is in its raw form and can be directly executed by postgres.","ref":"Torus.QueryInspector.html#tap_substituted_sql/3"},{"type":"function","title":"Example - Torus.QueryInspector.tap_substituted_sql/3","doc":"iex> Post\n    ...> |> where(id: 1)\n    ...> # ... Your complex query\n    ...> |> Torus.QueryInspector.tap_substituted_sql(Torus.Test.Repo)\n    ...> |> Repo.all()","ref":"Torus.QueryInspector.html#tap_substituted_sql/3-example"},{"type":"behaviour","title":"Torus.Embedding","doc":"Behaviour for generating embeddings for a list of terms.","ref":"Torus.Embedding.html"},{"type":"callback","title":"Torus.Embedding.embedding_model/1","doc":"Returns a string representing the embedding model name `Torus` currently uses to generate embeddings.\n\nFor example: `\"sentence-transformers/paraphrase-MiniLM-L6-v2\"`.","ref":"Torus.Embedding.html#c:embedding_model/1"},{"type":"callback","title":"Torus.Embedding.generate/2","doc":"Generates embeddings for a given list of terms.\n\nShould raise or retry on errors.","ref":"Torus.Embedding.html#c:generate/2"},{"type":"module","title":"Torus.Embeddings.Batcher","doc":"Size/time‑bounded **batcher** for embedding generation.\n\n`Torus.Embeddings.Batcher` is a long‑running GenServer that collects\nindividual `generate/2` calls, groups them into a single batch, and forwards the\nbatch to the configured `embedding_module`.\n\n---","ref":"Torus.Embeddings.Batcher.html"},{"type":"module","title":"Why batch? - Torus.Embeddings.Batcher","doc":"* **Fewer model / network invocations** – one request with *n* terms is cheaper\nthan *n* single‑term requests.\n* **Lower latency under load** – callers wait only for the current batch to\nflush, not for an entire queue of independent requests.\n* **Higher throughput per API quota** – most providers charge per request, so\nbatched calls extract more value from the same quota.\n\n---","ref":"Torus.Embeddings.Batcher.html#module-why-batch"},{"type":"module","title":"Flush conditions - Torus.Embeddings.Batcher","doc":"A batch is flushed when **either** condition is met (whichever comes first):\n\n* the queue reaches `max_batch_size` terms, or\n* `max_batch_wait_ms` elapses after the first term was queued.\n\nBoth limits are fully configurable.\n\n---","ref":"Torus.Embeddings.Batcher.html#module-flush-conditions"},{"type":"module","title":"Configuration - Torus.Embeddings.Batcher","doc":"It's considered a good practise to batch requests to the embedding module, especially when you are dealing with a high-traffic applications.\n\nTo use it:\n\n- Add the following to your `config.exs`:\n\n```elixir\n config :torus, batcher: Torus.Embeddings.Batcher\n\n config :torus, Torus.Embeddings.Batcher,\n    max_batch_size: 10,\n    default_batch_timeout: 100,\n    embedding_module: Torus.Embeddings.HuggingFace\n```\n\n- Add it to your supervision tree:\n\n```elixir\ndef start(_type, _args) do\n  children = [\n    # Your deps\n    Torus.Embeddings.Batcher\n  ]\n\n  opts = [strategy: :one_for_one, name: YourApp.Supervisor]\n  Supervisor.start_link(children, opts)\nend\n```\n\n- Configure your `embedding_module` of choice (see corresponding section)\n\nAnd you should be good to call `Torus.to_vector/1` and `Torus.to_vectors/1` functions.\n\nAlso, you can configure `call_timeout` option in `Torus.to_vector/2` and `Torus.to_vectors/2` functions to override the default timeout for the batching call. This is useful if you're okay to wait longer for the batch to flush and your embedder to generate the embedding.\n\nSee `Torus.semantic/5` on how to use this module to introduce semantic search in your application.","ref":"Torus.Embeddings.Batcher.html#module-configuration"},{"type":"function","title":"Torus.Embeddings.Batcher.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Torus.Embeddings.Batcher.html#child_spec/1"},{"type":"function","title":"Torus.Embeddings.Batcher.embedding_model/1","doc":"","ref":"Torus.Embeddings.Batcher.html#embedding_model/1"},{"type":"function","title":"Torus.Embeddings.Batcher.generate/2","doc":"","ref":"Torus.Embeddings.Batcher.html#generate/2"},{"type":"function","title":"Torus.Embeddings.Batcher.init/1","doc":"","ref":"Torus.Embeddings.Batcher.html#init/1"},{"type":"function","title":"Torus.Embeddings.Batcher.start_link/1","doc":"","ref":"Torus.Embeddings.Batcher.html#start_link/1"},{"type":"module","title":"Torus.Embeddings.HuggingFace","doc":"A wrapper around Hugging Face API. It allows you to generate embeddings using a variety of models available on Hugging Face.\n\nTo use it:\n\n- Add the following to your `config.exs`:\n\n  ```elixir\n  config :torus, embedding_module: Torus.Embeddings.HuggingFace\n  ```\n\n- Add `req` to your `mix.exs` dependencies:\n\n  ```elixir\n  def deps do\n  [\n    {:req, \"~> 0.5\"}\n  ]\n  end\n  ```\n\n- Add an API token for hugging face to your `runtime.exs`. You can get your token [here](https://huggingface.co/settings/tokens).\n\n  ```elixir\n  config :torus, Torus.Embeddings.HuggingFace, token: System.get_env(\"HUGGING_FACE_API_KEY\")\n  ```\n\nBy default, it uses `sentence-transformers/all-MiniLM-L6-v2` model, but you can specify a different model by explicitly passing `model` to the config:\n\n```elixir\nconfig :torus, Torus.Embeddings.HuggingFace, model: \"your/model\"\n```\n\nSee `Torus.semantic/5` on how to use this module to introduce semantic search in your application.","ref":"Torus.Embeddings.HuggingFace.html"},{"type":"module","title":"Torus.Embeddings.LocalNxServing","doc":"Local embedding generator using local Hugging Face model with Bumblebee and Nx.Serving.\n\nIt allows you to generate embeddings on your local machine using a variety of models available on Hugging Face.\n\nTo use it:\n\n- Add the following to your `config.exs`:\n\n  ```elixir\n  config :torus, embedding_module: Torus.Embeddings.LocalNxServing\n  ```\n\n- Add `bumblebee` and `nx` to your `mix.exs` dependencies:\n\n  ```elixir\n  def deps do\n  [\n    {:bumblebee, \"~> 0.6\"},\n    {:nx, \"~> 0.9\"}\n  ]\n  end\n  ```\n\n- Add it to your supervision tree:\n\nHere you'd probably want to start it only on machines with GPU. See more info in [Nx Serving documentation](https://hexdocs.pm/nx/Nx.Serving.html)\n\n```elixir\ndef start(_type, _args) do\n  children = [\n    # Your deps\n    Torus.Embeddings.LocalNxServing\n  ]\n\n  opts = [strategy: :one_for_one, name: YourApp.Supervisor]\n  Supervisor.start_link(children, opts)\nend\n```\n\nYou can pass all options directly to `Nx.Serving.start_link/1` function by passing them to `Torus.Embeddings.LocalNxServing` when starting.\n\nBy default, it uses `sentence-transformers/all-MiniLM-L6-v2` model, but you can specify a different model by explicitly passing `model` to the config:\n\n```elixir\nconfig :torus, Torus.Embeddings.LocalNxServing, model: \"your/model\"\n```\n\nSee `Torus.semantic/5` on how to use this module to introduce semantic search in your application.","ref":"Torus.Embeddings.LocalNxServing.html"},{"type":"function","title":"Torus.Embeddings.LocalNxServing.child_spec/1","doc":"","ref":"Torus.Embeddings.LocalNxServing.html#child_spec/1"},{"type":"function","title":"Torus.Embeddings.LocalNxServing.start_link/1","doc":"","ref":"Torus.Embeddings.LocalNxServing.html#start_link/1"},{"type":"module","title":"Torus.Embeddings.NebulexCache","doc":"Caching layer for the embedding module.\n\nA wrapper around [Nebulex](https://hexdocs.pm/nebulex/Nebulex.html) cache. It allows you to cache the embedding calls in memory, so you save the resources/cost of calling the embedding module multiple times for the same input.\n\nTo use it:\n\n- Add the following to your `config.exs`:\n\n  ```elixir\n  config :torus, cache: Torus.Embeddings.NebulexCache\n  config :torus, Torus.Embeddings.NebulexCache,\n    embedding_module: Torus.Embeddings.PostgresML\n    cache: Nebulex.Cache,\n    otp_name: :your_app,\n    adapter: Nebulex.Adapters.Local,\n    # Other adapter-specific options\n    allocated_memory: 1_000_000_000, # 1GB\n\n  # Embedding module specific options\n  config :torus, Torus.Embeddings.PostgresML, repo: TorusExample.Repo\n  ```\n\n- Add `nebulex` and `decorator` to your `mix.exs` dependencies:\n\n  ```elixir\n  def deps do\n  [\n    {:nebulex, \">= 0.0.0\"},\n    {:decorator, \">= 0.0.0\"}\n  ]\n  end\n  ```\n\n- Add `Torus.Embeddings.NebulexCache` to your supervision tree:\n\n  ```elixir\n  def start(_type, _args) do\n    children = [\n      Torus.Embeddings.NebulexCache\n    ]\n\n    opts = [strategy: :one_for_one, name: YourApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\n  ```\n\nSee the [Nebulex documentation](https://hexdocs.pm/nebulex/Nebulex.html) for more information on how to configure the cache.\n\nAnd you're good to go now. As you can see, we can create a chain of embedding modules to compose the embedding process of our choice.\nEach of them should implement the `Torus.Embedding` behaviour, and we're good to go!","ref":"Torus.Embeddings.NebulexCache.html"},{"type":"function","title":"Torus.Embeddings.NebulexCache.generations/0","doc":"A convenience function for retrieving the current generations.","ref":"Torus.Embeddings.NebulexCache.html#generations/0"},{"type":"function","title":"Torus.Embeddings.NebulexCache.new_generation/1","doc":"A convenience function for creating new generations.","ref":"Torus.Embeddings.NebulexCache.html#new_generation/1"},{"type":"function","title":"Torus.Embeddings.NebulexCache.newer_generation/0","doc":"A convenience function for retrieving the newer generation.","ref":"Torus.Embeddings.NebulexCache.html#newer_generation/0"},{"type":"function","title":"Torus.Embeddings.NebulexCache.reset_generation_timer/0","doc":"A convenience function for reset the GC timer.","ref":"Torus.Embeddings.NebulexCache.html#reset_generation_timer/0"},{"type":"module","title":"Torus.Embeddings.OpenAI","doc":"A wrapper around OpenAI API. It allows you to generate embeddings using OpenAI models.\n\nTo use it:\n\n- Add the following to your `config.exs`:\n\n  ```elixir\n  config :torus, embedding_module: Torus.Embeddings.OpenAI\n  ```\n\n- Add `req` to your `mix.exs` dependencies:\n\n  ```elixir\n  def deps do\n  [\n    {:req, \"~> 0.5\"}\n  ]\n  end\n  ```\n\n- Add an API token for hugging face to your `runtime.exs`. You can get your token [here](https://huggingface.co/settings/tokens).\n\n  ```elixir\n  config :torus, Torus.Embeddings.OpenAI, token: System.get_env(\"OPEN_AI_API_KEY\")\n  ```\n\nBy default, it uses `sentence-transformers/all-MiniLM-L6-v2` model, but you can specify a different model by explicitly passing `model` to the config:\n\n```elixir\nconfig :torus, Torus.Embeddings.OpenAI, model: \"your/model\"\n```\n\nSee `Torus.semantic/5` on how to use this module to introduce semantic search in your application.","ref":"Torus.Embeddings.OpenAI.html"},{"type":"module","title":"Torus.Embeddings.PostgresML","doc":"`Torus.Embeddings.PostgresML` uses PostgreSQL [PostgresML extension](https://PostgresML.org/docs) to generate embeddings. It allows you to generate embeddings using a variety of models and performs inference directly in the database. This would require your database to have GPU support.\n\nTo use it, add the following to your `config.exs`:\n\n```elixir\nconfig :torus, embedding_module: Torus.Embeddings.PostgresML\nconfig :torus, Torus.Embeddings.PostgresML, repo: YourApp.Repo\n```\n\nBy default, it uses `sentence-transformers/all-MiniLM-L6-v2` model, but you can specify a different model by explicitly passing `model` to the config:\n\n```elixir\nconfig :torus, Torus.Embeddings.PostgresML, model: \"your/model\"\n```\n\nRead more about in [PostgresML](https://PostgresML.org/blog/semantic-search-in-postgres-in-15-minutes).\n\nSee `Torus.semantic/5` on how to use this module to introduce semantic search in your application.","ref":"Torus.Embeddings.PostgresML.html"},{"type":"extras","title":"Semantic search","doc":"# Semantic search\n\nThis guide will walk you through adding semantic search to your application using `Torus` and PostgreSQL.","ref":"semantic_search.html"},{"type":"extras","title":"Initial requirements - Semantic search","doc":"Your app needs to use PostgreSQL and Ecto, and you should be comfortable storing and comparing embedded vectors in PostgreSQL.","ref":"semantic_search.html#initial-requirements"},{"type":"extras","title":"Getting started - Semantic search","doc":"Semantic search is a technology that understands the meaning behind words to deliver results that match a user's intent, not just exact keywords. It uses AI and machine learning to interpret natural language in context, improving the relevance of search results.\n\n**An embedding** is a numerical representation (vector) of a word, sentence, or document in a high‑dimensional space. It captures the semantic meaning of the text, allowing for more accurate comparisons and searches. The more dimensions an embedding has, the more information it can capture. However, higher dimensions also mean greater complexity and computational cost.","ref":"semantic_search.html#getting-started"},{"type":"extras","title":"The search process is split into three phases: - Semantic search","doc":"1. Generate and store the embeddings for the data you want to search.\n\n   We'll use `Torus.to_vectors/1` function. We'll dive deeper into how to do this efficiently later.\n\n2. Generate an embedding for the search term.\n\n   This is also done with the `Torus.to_vector/1` function.\n\n3. Compare the embedding of the search term with the embeddings of the data you want to search.\n\n   We'll do this using `Torus.semantic/5` function.\n\nOverall, our search will look like this:\n\n```elixir\ndef search(term) do\n  # Generate an embedding for the search term\n  search_vector = Torus.to_vector(term)\n\n  Post\n  |> Torus.semantic([p], p.embedding, search_vector)\n  |> Repo.all()\nend\n```\n\nNote: You’ll need to join or preload the associated embeddings if they're stored in a separate table.","ref":"semantic_search.html#the-search-process-is-split-into-three-phases"},{"type":"extras","title":"1. Generating embeddings - Semantic search","doc":"There are several ways to generate embeddings. `Torus` includes a set of built‑in `Torus.Embeddings` modules that implement the `Torus.Embedding` behaviour - but you're not limited to those. You can easily implement the `Torus.Embedding` behaviour yourself — it’s designed to be simple and straightforward.\n\nHere is what `Torus` provides out of the box:","ref":"semantic_search.html#1-generating-embeddings"},{"type":"extras","title":"Torus.Embeddings.HuggingFace - Semantic search","doc":"`Torus.Embeddings.HuggingFace` is a wrapper around the Hugging Face API. It allows you to generate embeddings using a variety of models available on Hugging Face.\n\nTo use it:\n\n- Add the following to your `config.exs`:\n\n  ```elixir\n  config :torus, embedding_module: Torus.Embeddings.HuggingFace\n  ```\n\n- Add `req` to your `mix.exs` dependencies:\n\n  ```elixir\n  def deps do\n  [\n    {:req, \"~> 0.5\"}\n  ]\n  end\n  ```\n\n- Add an API token for Hugging Face to your `runtime.exs`. You can get your token [here](https://huggingface.co/settings/tokens).\n\n  ```elixir\n  config :torus, Torus.Embeddings.HuggingFace, token: System.get_env(\"HUGGING_FACE_API_KEY\")\n  ```\n\nBy default, it uses the `sentence-transformers/all-MiniLM-L6-v2` model, but you can specify a different model by explicitly passing `model` in the configuration:\n\n```elixir\nconfig :torus, Torus.Embeddings.HuggingFace, model: \"your/model\"\n```","ref":"semantic_search.html#torus-embeddings-huggingface"},{"type":"extras","title":"Torus.Embeddings.OpenAI - Semantic search","doc":"`Torus.Embeddings.OpenAI` is a wrapper around the OpenAI API. It allows you to generate embeddings using OpenAI models.\n\nTo use it:\n\n- Add the following to your `config.exs`:\n\n  ```elixir\n  config :torus, embedding_module: Torus.Embeddings.OpenAI\n  ```\n\n- Add `req` to your `mix.exs` dependencies:\n\n  ```elixir\n  def deps do\n  [\n    {:req, \"~> 0.5\"}\n  ]\n  end\n  ```\n\n- Add an API token for OpenAI to your `runtime.exs`. You can get your token [here](https://platform.openai.com/account/api-keys).\n\n  ```elixir\n  config :torus, Torus.Embeddings.OpenAI, token: System.get_env(\"OPEN_AI_API_KEY\")\n  ```\n\nBy default, it uses the `sentence-transformers/all-MiniLM-L6-v2` model, but you can specify a different model by explicitly passing `model` in the configuration:\n\n```elixir\nconfig :torus, Torus.Embeddings.OpenAI, model: \"your/model\"\n```","ref":"semantic_search.html#torus-embeddings-openai"},{"type":"extras","title":"Torus.Embeddings.PostgresML - Semantic search","doc":"`Torus.Embeddings.PostgresML` uses the PostgreSQL [PostgresML extension](https://postgresml.org/docs) to generate embeddings. It allows you to generate embeddings using a variety of models and performs inference directly in the database. This requires your database to have GPU support.\n\nTo use it, add the following to your `config.exs`:\n\n```elixir\nconfig :torus, embedding_module: Torus.Embeddings.PostgresML\nconfig :torus, Torus.Embeddings.PostgresML, repo: YourApp.Repo\n```\n\nBy default, it uses the `sentence-transformers/all-MiniLM-L6-v2` model, but you can specify a different model by explicitly passing `model` in the configuration:\n\n```elixir\nconfig :torus, Torus.Embeddings.PostgresML, model: \"your/model\"\n```","ref":"semantic_search.html#torus-embeddings-postgresml"},{"type":"extras","title":"Torus.Embeddings.LocalNxServing - Semantic search","doc":"`Torus.Embeddings.LocalNxServing` will probably require an instance with GPU support. It allows you to generate embeddings on your local machine using a variety of models available on Hugging Face. It leverages the `nx` and `bumblebee` libraries to perform inference.\n\nTo use it:\n\n- Add the following to your `config.exs`:\n\n  ```elixir\n  config :torus, embedding_module: Torus.Embeddings.LocalNxServing\n  ```\n\n- Add `bumblebee` and `nx` to your `mix.exs` dependencies:\n\n  ```elixir\n  def deps do\n  [\n    {:bumblebee, \"~> 0.6\"},\n    {:nx, \"~> 0.9\"}\n  ]\n  end\n  ```\n\n- Add it to your supervision tree:\n\n  Here you'd probably want to start it only on machines with GPUs. See more information in the [Nx.Serving documentation](https://hexdocs.pm/nx/Nx.Serving.html).\n\n  ```elixir\n  def start(_type, _args) do\n    children = [\n      # Your deps\n      Torus.Embeddings.LocalNxServing\n    ]\n\n    opts = [strategy: :one_for_one, name: YourApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\n  ```\n\nYou can pass all options directly to the `Nx.Serving.start_link/1` function by passing them to `Torus.Embeddings.LocalNxServing` when starting.\n\nBy default, it uses the `sentence-transformers/all-MiniLM-L6-v2` model, but you can specify a different model by explicitly passing `model` in the configuration:\n\n```elixir\nconfig :torus, Torus.Embeddings.LocalNxServing, model: \"your/model\"\n```","ref":"semantic_search.html#torus-embeddings-localnxserving"},{"type":"extras","title":"Torus.Embeddings.Batcher - Semantic search","doc":"`Torus.Embeddings.Batcher` is a long‑running **GenServer** that collects individual embedding calls, groups them into a single batch, and forwards the batch to the configured `embedding_module`. It can be used in combination with any of the embedding modules above.\n\nIt is considered good practice to batch requests to the embedding module, especially when you are dealing with high‑traffic applications.\n\nTo use it:\n\n- Add the following to your `config.exs`:\n\n  ```elixir\n  config :torus, batcher: Torus.Embeddings.Batcher\n\n  config :torus, Torus.Embeddings.Batcher,\n    max_batch_size: 10,\n    default_batch_timeout: 100,\n    embedding_module: Torus.Embeddings.HuggingFace\n  ```\n\n- Add it to your supervision tree:\n\n  ```elixir\n  def start(_type, _args) do\n    children = [\n      # Your deps\n      Torus.Embeddings.Batcher\n    ]\n\n    opts = [strategy: :one_for_one, name: YourApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\n  ```\n\n- Configure your `embedding_module` of choice (see the corresponding section).\n\nAfter that, you can call the `Torus.to_vector/1` and `Torus.to_vectors/1` functions.","ref":"semantic_search.html#torus-embeddings-batcher"},{"type":"extras","title":"Torus.Embeddings.NebulexCache - Semantic search","doc":"`Torus.Embeddings.NebulexCache` is a wrapper around the [Nebulex](https://hexdocs.pm/nebulex/readme.html) cache. It allows you to cache embedding calls in memory, so you save resources and cost by avoiding repeated calls to the embedding module for the same input.\n\nTo use it:\n\n- Add the following to your `config.exs`:\n\n  ```elixir\n  config :torus, cache: Torus.Embeddings.NebulexCache\n  config :torus, Torus.Embeddings.NebulexCache,\n    embedding_module: Torus.Embeddings.PostgresML,\n    cache: Nebulex.Cache,\n    otp_name: :your_app,\n    adapter: Nebulex.Adapters.Local,\n    # Other adapter‑specific options\n    allocated_memory: 1_000_000_000 # 1 GB\n\n  # Embedding‑module‑specific options\n  config :torus, Torus.Embeddings.PostgresML, repo: TorusExample.Repo\n  ```\n\n- Add `nebulex` and `decorator` to your `mix.exs` dependencies:\n\n  ```elixir\n  def deps do\n  [\n    {:nebulex, \">= 0.0.0\"},\n    {:decorator, \">= 0.0.0\"}\n  ]\n  end\n  ```\n\n- Add `Torus.Embeddings.NebulexCache` to your supervision tree:\n\n  ```elixir\n  def start(_type, _args) do\n    children = [\n      Torus.Embeddings.NebulexCache\n    ]\n\n    opts = [strategy: :one_for_one, name: YourApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\n  ```\n\nSee the [Nebulex documentation](https://hexdocs.pm/nebulex/Nebulex.html) for more information on how to configure the cache.\n\nAnd you're good to go. As you can see, you can create a chain of embedding modules to compose an embedding process of your choice. Each of them should implement the `Torus.Embedding` behaviour, and you're all set!","ref":"semantic_search.html#torus-embeddings-nebulexcache"},{"type":"extras","title":"2. Storing the embeddings - Semantic search","doc":"","ref":"semantic_search.html#2-storing-the-embeddings"},{"type":"extras","title":"Database structure - Semantic search","doc":"There are multiple ways to store embeddings, but I'd recommend the following approach:\n\n```elixir\ncreate table(:embeddings) do\n  # Model name as a string so we can differentiate between models\n  add :model, :string, null: false\n  # Parameters used to generate the embedding (maybe an embedding version) so we can\n  # filter by the newest later\n  add :metadata, :jsonb, null: false, default: \"{}\"\n  # Maybe more columns to differentiate this embedding from others?\n  # Actual embedding vector (384 dimensions for the `all-MiniLM-L6-v2` model)\n  add :embedding, :vector, size: 384, null: false\nend\n```\n\n```elixir\ncreate table(:posts) do\n  add :title, :string\n  add :body, :string\n\n  timestamps(type: :utc_datetime)\nend\n```\n\n```elixir\ncreate table(:post_embeddings) do\n  add :post_id, references(:posts, on_delete: :delete_all), null: false\n  add :embedding_id, references(:embeddings, on_delete: :delete_all), null: false\nend\n```\n\nThis gives us a many‑to‑many relationship between posts and embeddings, so we can version the embeddings and even store multiple embeddings for the same post. However, I'd suggest concatenating all string/binary fields together to generate one embedding per post for simpler searches later on.\n\nThere is also a pattern to store a vector without the dimensions and its dimensions and metadata in separate columns, but I'd not recommend it since we won't be able to [index this vector](https://tembo.io/blog/vector-indexes-in-pgvector) later on.","ref":"semantic_search.html#database-structure"},{"type":"extras","title":"The embedding process – existing rows - Semantic search","doc":"I suggest inserting an [Oban](https://hexdocs.pm/oban/Oban.html) job to generate embeddings in chunks for all rows in the database, using the `Torus.to_vectors/1` function.","ref":"semantic_search.html#the-embedding-process-existing-rows"},{"type":"extras","title":"The embedding process – new rows - Semantic search","doc":"<!-- TODO: Add Oban job helper that can help the embedding process -->\n\nThere are a few ways to handle the embedding process for new rows:\n\n1. Add a cron Oban job to run periodically and, in batches, embed all rows that need it.\n2. Schedule an Oban job to embed the row after inserting it into the database.\n3. (Least recommended) Embed the row in the same transaction used to insert it into the database.","ref":"semantic_search.html#the-embedding-process-new-rows"},{"type":"extras","title":"3. Searching - Semantic search","doc":"We need to generate the embedding and then compare it with the embeddings in the database. This can be done using the `Torus.semantic/5` function.\n\n```elixir\ndef search(term) do\n  search_vector = Torus.to_vector(term)\n\n  Post\n  |> Torus.semantic([p], p.embedding, search_vector, distance: :l2_distance, pre_filter: 0.7)\n  |> Repo.all()\nend\n```","ref":"semantic_search.html#3-searching"},{"type":"extras","title":"Changelog","doc":"# v0.5.0\n\n- Similarity search type now defaults to `:word_similarity` instead of `similarity`. \n- Possible `similarity/5` search types are updated to be prefixed with `similarity` to replicate 1-1 these in `pg_trgm` extension.\n- Extended optimization section in the docs\n\n# v0.4.1\n\nMinor doc updates\n\n# v0.4.0","ref":"changelog.html"},{"type":"extras","title":"Breaking changes: - Changelog","doc":"- `full_text/5` - now returns all results when search term contains a stop word or is empty instead of returning none.","ref":"changelog.html#breaking-changes"},{"type":"extras","title":"Improvements: - Changelog","doc":"- `full_text/5` - now supports `:empty_return` option that controls if the query should return all results when search term contains a stop word or is empty.\n- `tap_explain_analyze/3` - now correctly returns the query plan.\n- Docs were grouped together by the search type.","ref":"changelog.html#improvements"},{"type":"extras","title":"New 🔥 - Changelog","doc":"**Semantic search** is finally here! Read more about it in the [Semantic search with Torus](/guides/semantic_search.md) guide.\nShortly - it allows you to generate embeddings using a configurable adapters and use them to compare against the ones stored in your database.\n\nSupported adapters (for now):\n\n- `Torus.Embeddings.OpenAI` - uses OpenAI's API to generate embeddings.\n\n- `Torus.Embeddings.HuggingFace` - uses HuggingFace's API to generate embeddings.\n\n- `Torus.Embeddings.LocalNxServing` - generate embeddings on your local machine using a variety of models available on Hugging Face\n\n- `Torus.Embeddings.PostgresML` - uses PostgreSQL [PostgresML extension](https://PostgresML.org/docs) to generate embeddings\n\n- `Torus.Embeddings.Batcher` - a long‑running **GenServer** that collects individual embedding calls, groups them into a single batch, and forwards the batch to the configured `embedding_module` (any from the above or your custom one).\n\n- `Torus.Embeddings.NebulexCache` - a wrapper around [Nebulex](https://hexdocs.pm/nebulex/readme.html) cache, allowing you to cache the embedding calls in memory, so you save the resources/cost of calling the embedding module multiple times for the same input.\n\nAnd you can easily create your own adapter by implementing the `Torus.Embedding` behaviour.\n\n# v0.3.0","ref":"changelog.html#new"},{"type":"extras","title":"Breaking changes: - Changelog","doc":"- `full_text_dynamic/5` is renamed to `full_text/5` and now supports stored columns.\n- `similarity/5` - `limit` option is removed, use Ecto's `limit/2` instead.\n- `full_text/5` - `:concat` option is renamed to `:coalesce`.","ref":"changelog.html#breaking-changes"},{"type":"extras","title":"Improvements: - Changelog","doc":"- `full_text/5` now supports stored `tsvector` columns.\n- `Torus.QueryInspector.substituted_sql/3` now correctly handles arrays substitutions.\n- Docs are extended to guide through the performance and relevance.\n\nAnd other minor performance/clearance improvements.\n\n# v0.2.2\n\n- `full_text_dynamic/5`: Replaced `:nullable_columns` with `:concat` option\n- `similarity/5`: Fixed a bug where you weren't able to pass variable as a term\n- `Torus.QueryInspector`: now is not tied with `Torus.Testing` and serves as a separate standalone module.\n\nAnd other minor performance/clearance improvements.\n\n# v0.2.1\n\n`similarity/5` search is now fully tested and customizable. `full_text_dynamic/5` is up next.\n\n# Changelog for Torus v0.2.0\n\nTorus now supports full text search, ilike, and similarity search.","ref":"changelog.html#improvements"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.37.3"}}